#include <math.h>

#include "main.h"

/**
 * \brief Vypis napovedy
 *
 * Vypise do terminalu zpusob pouziti programu
 * @param str Nazev programu - obvykle argv[0]
 */
void pouziti(gchar *str)
{
	g_print("Hammicoder %s\n", VERZE);
	g_print("Pouziti:\n%s [-edg] [-v] [-N NATIVE] [-H HAMMING]\n", str);
	g_print("\t-e ....... encodovani\n");
	g_print("\t-d ....... dekodovani\n");
	g_print("\t-g ....... graficke rozhrani (vychozi)\n");
	g_print("\t-v ....... upovidany rezim\n");
	//g_print("\t-V ....... debugovaci vypisy\n");	// Pri velkych souborech extremne pomale - radsi neukazovat :-)
	g_print("\tNATIVE ... retezec nativniho kodu\n");
	g_print("\tHAMMING .. retezec hammingova kodu\n");
}

/**
 * \brief Mereni rychlosti a prubehu
 *
 * Meri a zobrazuje rychlost komprese v MB/s a prubeh zpracovani v %
 * @param data Struktura obsahujici pointery na graficke prvky
 * @return TRUE pokud vypocet pokracuje, FALSE po ukonceni
 */
gboolean merak_cb(gpointer data)
{
	glong diff = pro_merak - pro_merak_posledni;
	gfloat speed;
	ChData *chdata = (ChData*) data;
	gchar buf[50];
	gfloat prubeh;

	pro_merak_posledni = pro_merak;

	prubeh = (gfloat) pro_merak / (gfloat) pro_merak_celkem * 100;

	if (diff == 0) {
		if (grafika) gtk_label_set_label(GTK_LABEL(chdata->label_info), "Ukladam soubor");
		if (grafika) gtk_statusbar_push(GTK_STATUSBAR(chdata->statusbar), gtk_statusbar_get_context_id(GTK_STATUSBAR(chdata->statusbar), "uvitani"), VERZE);

		return FALSE;
	} else {
		speed = (gfloat) diff * 1000.0 / (gfloat) TIMEOUT / 1e6;
		if (vypisy > 1) g_print("%s\n", buf);
		sprintf(buf, "Zpracovano: %.1f%%", prubeh);
		if (grafika) gtk_label_set_label(GTK_LABEL(chdata->label_info), buf);
		sprintf(buf, "Aktualni rychlost: %.3fMB/s", speed);
		if (grafika) gtk_statusbar_push(GTK_STATUSBAR(chdata->statusbar), gtk_statusbar_get_context_id(GTK_STATUSBAR(chdata->statusbar), "rychlost"), buf);
		return TRUE;
	}
}

gchar xor(gchar a, gchar b)
{
	return(a == b) ? ('0') : ('1');
}

int encode(gpointer data)
{
	gchar *buf;
	gchar *ham;
	GtkTextIter start, end;
	ChData *chdata = (ChData*) data;
	gint i, j;
	gint delka;
	gchar delka_parit;
	gchar offset;
	gchar mocnina;
	int tmp;

	gtk_label_set_text(GTK_LABEL(chdata->label_info), "Enkoduji");

	gtk_text_buffer_get_start_iter(GTK_TEXT_BUFFER(chdata->textbuffer1), &start);
	gtk_text_buffer_get_end_iter(GTK_TEXT_BUFFER(chdata->textbuffer1), &end);
	delka = gtk_text_buffer_get_char_count(GTK_TEXT_BUFFER(chdata->textbuffer1));
	buf = malloc(sizeof(gchar) * delka);
	buf = gtk_text_buffer_get_text(GTK_TEXT_BUFFER(chdata->textbuffer1), &start, &end, TRUE);
	delka_parit = log2(delka);
	ham = malloc(sizeof(gchar)*(delka + delka_parit + 5)); /**< Radi trochu vetsi - mozno upravit */
	memset(ham, '0', delka + delka_parit + 5);

	if (vypisy > 0) g_print("Enkoduji\n%s\n", buf);
	for (i = 0; i < delka; i++) {
		if (buf[i] != '0' && buf[i] != '1') {
			gtk_label_set_text(GTK_LABEL(chdata->label_info), "Jen znaky 0, 1!");
			return 1;
		}
	}
	offset = 2;
	mocnina = 2;
	for (i = 0; i < delka; i++) {
		tmp = i + offset; /**< Pro pouziti u vypoctu parit */
		ham[tmp] = buf[i]; /**< nezmenena data ze vstupu */
		j = 0;
		if (vypisy > 0) g_print("i=%d offset=%d tmp=%d mocnina=%d\n", i, offset, tmp, mocnina);
		while (j < mocnina) {
			if (vypisy > 0) g_print("j=%d x=%d >=? y=%d\n", j, (tmp + 1) % (gint) (exp2(j + 1)), (gint) (exp2(j)));
			if ((tmp + 1) % (gint) (exp2(j + 1)) >= (gint) (exp2(j))) {
				if (vypisy > 0) g_print("Xoruju.\n");
				ham[(gint) (exp2(j)) - 1] = xor(ham[(gint) (exp2(j)) - 1], buf[i]);
			}
			j++;
		}
		if (i != delka - 1) { /**< Pro zamezeni inkrementu offsetu po zapisu posledniho bitu */
			if (tmp + 1 == (gint) (exp2(mocnina) - 1)) { /**< nasleduje pozice paritniho bitu */
				mocnina++;
				offset++;
			}
		}
	}
	for (j = 0; j < i + offset; j++) {
		ham[i + offset] = xor(ham[i + offset], ham[j]);
	}

	ham[i + offset + 1] = '\0'; /**< Aby se choval ham jako string */
	if (vypisy > 0) g_print("Vystup\n%s\n", ham);

	gtk_text_buffer_set_text(GTK_TEXT_BUFFER(chdata->textbuffer2), ham, -1);
	//gtk_text_buffer_set_text(GTK_TEXT_BUFFER(chdata->textbuffer1), buf, -1);

	free(buf);
	free(ham);

	gtk_label_set_text(GTK_LABEL(chdata->label_info), "Hotovo");

	return 0;
}

int decode(gpointer data)
{
	gchar *buf;
	gchar *nat;
	GtkTextIter start, end;
	ChData *chdata = (ChData*) data;
	gint i, j;
	gint priznakSx; /**< Priznak Sx = pozice chybneho bitu (pri jedne chybe) */
	gchar priznakS;
	gint delka;
	gchar *parity;
	gchar delka_parit;
	gchar offset;
	gchar mocnina;
	gchar tmp[50];
	gint pozice;

	gtk_label_set_text(GTK_LABEL(chdata->label_info), "Dekoduji");

	gtk_text_buffer_get_start_iter(GTK_TEXT_BUFFER(chdata->textbuffer2), &start);
	gtk_text_buffer_get_end_iter(GTK_TEXT_BUFFER(chdata->textbuffer2), &end);
	delka = gtk_text_buffer_get_char_count(GTK_TEXT_BUFFER(chdata->textbuffer2));
	delka_parit = (gint) (log2(delka) + 5);
	parity = malloc(delka_parit); /**< Stacilo by mensi - mozno upravit */
	buf = malloc(sizeof(gchar) * delka);
	buf = gtk_text_buffer_get_text(GTK_TEXT_BUFFER(chdata->textbuffer2), &start, &end, TRUE);
	nat = malloc(sizeof(gchar) * delka); /**< Stacilo by mensi - mozno upravit */
	memset(nat, '\0', delka);

	if (vypisy > 0) g_print("Dekoduji\n%s\n", buf);
	for (i = 0; i < delka; i++) {
		if (buf[i] != '0' && buf[i] != '1') {
			gtk_label_set_text(GTK_LABEL(chdata->label_info), "Jen znaky 0, 1!");
			return 1;
		}
	}

	for (i = 0; i < delka_parit; i++) {
		parity[i] = '0';
	}

	priznakS = '0';
	offset = 2;
	mocnina = 0;
	pozice = 0;
	for (i = 0; i < delka - 1; i++) {
		priznakS = xor(priznakS, buf[i]); /**< Celkova parita */
		if (i == (gint) (exp2(mocnina) - 1)) { /**< Pozice parit */
			parity[mocnina] = buf[i]; /**< Diky tomuto lze vynechat inicializaci ve foru vyse */
			mocnina++;
		} else {
			nat[pozice++] = buf[i]; /**< Ulozeni nativniho bitu */
			j = 0;
			while (j < mocnina) { /**< Prepocet paritnich bitu */
				if (vypisy > 0) g_print("j=%d x=%d >=? y=%d\n", j, (i + 1) % (gint) (exp2(j + 1)), (gint) (exp2(j)));
				if ((i + 1) % (gint) (exp2(j + 1)) >= (gint) (exp2(j))) {
					if (vypisy > 0) g_print("Xoruju.\n");
					parity[j] = xor(parity[j], buf[i]);
				}
				j++;
			}
		}
	}
	priznakS = xor(priznakS, buf[delka-1]);
	priznakSx = 0; /**< Pozice chyby (pri vyskytu jedne chyby) */
	for (i = 0; i < mocnina; i++) {
		if (parity[i] == '1')
			priznakSx += (gint) (exp2(i));
	}
	if (vypisy > 0) g_print("Sx = %d S = %c\n", priznakSx, (gchar) priznakS);

	if (priznakS == '0' && priznakSx == 0) {
		gtk_text_buffer_set_text(GTK_TEXT_BUFFER(chdata->textbuffer1), nat, -1);
		gtk_label_set_text(GTK_LABEL(chdata->label_info), "Bez chyb");
	} else {
		if (priznakS == '0') /**< priznakSx != 0 */
			gtk_label_set_text(GTK_LABEL(chdata->label_info), "Neopravitelna chyba");
		else if (priznakSx == 0) /**< priznakS != 0 */
			gtk_label_set_text(GTK_LABEL(chdata->label_info), "2n nasobna chyba");
		else { /**< priznakS != 0 a taky priznakSx != 0 */
			sprintf(tmp, "Chyba %d. bitu", priznakSx);
			gtk_label_set_text(GTK_LABEL(chdata->label_info), tmp);
		}
	}

	free(buf);
	free(nat);
	free(parity);

	return 0;
}

/**
 * \brief Callback ukonceni programu
 *
 * Ukonci hlavni smycku GTK.
 * @param widget Pointer na volajici widget
 * @param data Struktura obsahujici pointery na graficke prvky
 */
G_MODULE_EXPORT void quit_cb(GtkWidget *widget, ChData *data)
{
	fprintf(stderr, "Koncim\n");
	gtk_main_quit();
}

/**
 * \brief Callback kliku na tlacitko komprese
 *
 * Nastavi informaci o spusteni komprese a spusti funkci komprese v novem
 * vlaknu.
 * @param widget Pointer na widget tlacitka
 * @param data Struktura obsahujici pointery na graficke prvky
 */
G_MODULE_EXPORT void encode_cb(GtkWidget *widget, ChData *data)
{
	if ((thread = g_thread_create((GThreadFunc) encode, data, FALSE, &err)) == NULL) {
		fprintf(stderr, "Spusteni threadu se nepovedlo: %s!\n", err->message);
		g_error_free(err);
	}
}

/**
 * \brief Callback pusteni klavesy v Native textview
 *
 * Spusti se pri zmacknuti kterekoli klavesy pri aktivnim textview. Vyuziva
 * se pro realtime aktualizaci vystupu.
 * @param widget
 * @param event
 * @param data
 * @return
 */
G_MODULE_EXPORT gboolean on_key_release_native(GtkWidget *widget, GdkEvent *event, ChData *data)
{
	/*
		if ((thread = g_thread_create((GThreadFunc) encode, data, FALSE, &err)) == NULL) {
			fprintf(stderr, "Spusteni threadu se nepovedlo: %s!\n", err->message);
			g_error_free(err);
		}
	 */
	encode(data);

	return TRUE;
}

/**
 * \brief Callback pusteni klavesy v Hamming textview
 *
 * Spusti se pri zmacknuti kterekoli klavesy pri aktivnim textview. Vyuziva
 * se pro realtime aktualizaci vystupu.
 * @param widget
 * @param event
 * @param data
 * @return
 */
G_MODULE_EXPORT gboolean on_key_release_hamming(GtkWidget *widget, GdkEvent *event, ChData *data)
{
	decode(data);

	return TRUE;
}

/**
 * \brief Callback kliku na tlacitko dekomprese
 *
 * Nastavi informaci o spusteni dekomprese a spusti funkci dekomprese v novem
 * vlaknu.
 * @param widget Pointer na widget tlacitka
 * @param data Struktura obsahujici pointery na graficke prvky
 */
G_MODULE_EXPORT void decode_cb(GtkWidget *widget, ChData *data)
{
	if ((thread = g_thread_create((GThreadFunc) decode, data, FALSE, &err)) == NULL) {
		fprintf(stderr, "Spusteni threadu se nepovedlo: %s!\n", err->message);
		g_error_free(err);
	}
}

/**
 * \brief Callback menu About
 *
 * Pomoci GtkBuilderu vytvori a zobrazi About dialog. Nastavi callback pro
 * zavreni okna (dulezite)
 * @param widget Pointer na volbu menu
 * @param data Struktura obsahujici pointery na graficke prvky
 */
G_MODULE_EXPORT void about_cb(GtkWidget *widget, ChData *data)
{
	GtkBuilder *builder;
	GtkWidget *about_dialog;
	GError *error = NULL;

	builder = gtk_builder_new();
	if (!gtk_builder_add_from_file(builder, ABOUT_FILE, &error)) {
		g_warning("%s", error->message);
		g_free(error);
		return;
	}
	about_dialog = GTK_WIDGET(gtk_builder_get_object(builder, "aboutdialog"));
	gtk_builder_connect_signals(builder, NULL);
	g_signal_connect_swapped(about_dialog, "response", G_CALLBACK(gtk_widget_destroy), about_dialog); // Aby fungovalo Close
	g_object_unref(G_OBJECT(builder));
	gtk_widget_show(about_dialog);
}

/**
 * \brief Hlavni funkce
 *
 * Provede nacteni parametru prikazove radky, inicializuje nazvy souboru a podle
 * voleb spusti kompresi, dekompresi, zobrazi napovedu nebo nacte a zobrazi
 * graficke rozhrani.
 * @param argc Pocet retezcu v shellu pri spusteni
 * @param argv Pole pointeru na retezce v shellu pri spusteni
 * @return 0 pokud nedojde k chybe
 */
gint main(int argc, char **argv)
{
	ChData *data; /**< Struktura obsahujici pointery na graficke prvky */
	GtkBuilder *builder; /**< Struktura pro nacteni GUI z .glade souboru */
	GError *error = NULL; /**< Chyby pri nacitani GUI */
	gint akce = GRAFIKA; /**< Prepinac provedene akce */
	gchar buf_native[1024] = {0}; /**< Retezec native */
	gchar buf_hamming[1024] = {0}; /**< Retezec hamming */
	gint c; /**< Promenna pouzita pro zpracovani parametru shellu */

	opterr = 0;
	grafika = 1;

	while ((c = getopt(argc, argv, "deghN:H:vV")) != -1) {
		switch (c) {
		case 'e':
			akce = ENCODE;
			grafika = 0;
			break;
		case 'd':
			akce = DECODE;
			grafika = 0;
			break;
		case 'g':
			akce = GRAFIKA;
			grafika = 1;
			break;
		case 'h':
			pouziti(argv[0]);
			return 0;
			break;
		case 'N':
			strcpy(buf_native, optarg);
			break;
		case 'H':
			strcpy(buf_hamming, optarg);
			break;
		case 'v':
			vypisy = 1;
			break;
		case 'V':
			vypisy = 2;
			break;
		case '?':
			if (optopt == 'i' || optopt == 'o')
				fprintf(stderr, "Parametru -%c chybi nazev souboru.\n", optopt);
			else if (isprint(optopt))
				fprintf(stderr, "Neznamy parametr '-%c'.\n", optopt);
			else
				fprintf(stderr, "Neznamy znak '\\x%x'.\n", optopt);
			return -1;
		default:
			return -1;
		}
	}

	if (akce == 0 && akce != GRAFIKA) {
		pouziti(argv[0]);
		return 0;
	}

	data = g_slice_new(ChData);
	data->textbuffer1 = (GtkTextBuffer*) buf_native;
	data->textbuffer2 = (GtkTextBuffer*) buf_hamming;

	if (akce == GRAFIKA) {
		gtk_init(&argc, &argv);
		g_thread_init(NULL);
		builder = gtk_builder_new();
		if (!gtk_builder_add_from_file(builder, GUI_FILE, &error)) {
			g_warning("%s", error->message);
			g_free(error);
			return 1;
		}

		data->main_window = GTK_WIDGET(gtk_builder_get_object(builder, "main_window"));
		data->label_info = GTK_WIDGET(gtk_builder_get_object(builder, "label_info"));
		data->textview_native = GTK_WIDGET(gtk_builder_get_object(builder, "text_vstup"));
		data->textview_hamming = GTK_WIDGET(gtk_builder_get_object(builder, "text_vystup"));
		data->statusbar = GTK_WIDGET(gtk_builder_get_object(builder, "statusbar"));
		data->textbuffer1 = gtk_text_view_get_buffer(GTK_TEXT_VIEW(data->textview_native));
		data->textbuffer2 = gtk_text_view_get_buffer(GTK_TEXT_VIEW(data->textview_hamming));

		gtk_text_buffer_set_text(GTK_TEXT_BUFFER(data->textbuffer1), buf_native, -1);
		gtk_text_buffer_set_text(GTK_TEXT_BUFFER(data->textbuffer2), buf_hamming, -1);

		gtk_builder_connect_signals(builder, data);
		g_object_unref(G_OBJECT(builder));
		gtk_widget_show(data->main_window);
		gtk_text_view_set_buffer(GTK_TEXT_VIEW(data->textview_native), data->textbuffer1);
		gtk_text_view_set_buffer(GTK_TEXT_VIEW(data->textview_hamming), data->textbuffer2);
		gtk_statusbar_push(GTK_STATUSBAR(data->statusbar), gtk_statusbar_get_context_id(GTK_STATUSBAR(data->statusbar), "uvitani"), VERZE);

		gtk_main();

	} else if (akce == ENCODE) {
		if (encode(data) != 0) {
			fprintf(stderr, "Kodovani se nezdarilo.\n");
			return -1;
		}

	} else if (akce == DECODE) {
		if (decode(data) != 0) {
			fprintf(stderr, "Dekodovani se nezdarilo.\n");
			return -1;
		}
	}

	g_slice_free(ChData, data);

	return 0;
}
